from langgraph.graph import END
from shy_sh.models import State
from shy_sh.agents.misc import has_tool_calls


def tool_calls_edge(state: State):
    last_message = state["tool_history"][-1]
    if has_tool_calls(last_message):
        return "tools"
    return END



from uuid import uuid4
from langchain_core.messages import ToolMessage, HumanMessage
from rich import print
from shy_sh.models import State
from shy_sh.agents.tools import tools_by_name
from shy_sh.settings import settings
from shy_sh.agents.misc import parse_react_tool


def tools_handler(state: State):
    last_message = state["tool_history"][-1]
    t_calls = _get_tool_calls(last_message)

    tool_answers = []
    for t_call in t_calls:
        try:
            t = tools_by_name[t_call["name"]]
            message = t.invoke(
                {
                    **t_call,
                    "args": {"state": state, **t_call["args"]},
                }
            )
        except Exception as e:
            print(f"[bold red]🚨 Tool error: {e}[/bold red]")
            message = ToolMessage(f"Tool error: {e}", tool_call_id=t_call["id"])

        if settings.llm.agent_pattern == "react":
            m = HumanMessage(content=f"Tool response:\n{message.content}")
            m.artifact = getattr(message, "artifact", None)
            message = m
        tool_answers.append(message)
    return {"tool_history": tool_answers}


def _get_react_tool_calls(message):
    react_tool = parse_react_tool(message)
    return [
        {
            "name": react_tool.tool,
            "args": {"arg": react_tool.arg},
            "id": uuid4().hex,
            "type": "tool_call",
        }
    ]


def _get_function_call_tool_calls(message):
    return message.tool_calls


def _get_tool_calls(message):
    match (settings.llm.agent_pattern):
        case "react":
            return _get_react_tool_calls(message)
        case "function_call":
            return _get_function_call_tool_calls(message)
        case _:
            raise ValueError("Unknown agent pattern")



import pyperclip
from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate
from shy_sh.agents.llms import get_llm
from shy_sh.utils import ask_confirm, syntax
from shy_sh.models import ToolMeta
from shy_sh.settings import settings
from textwrap import dedent
from rich.live import Live
from rich import print


msg_template = dedent(
    """
    The given task was {task}.
    Explain this {script_type} and why it should solve the task{lang_spec}.
    Be concise and please limit your explanation to the provided {script_type} and avoid suggesting alternative solutions or directly referencing the given task.

    ```
    {script}
    ```
    """
)


@chain
def explain_chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a shell expert. The current date and time is {timestamp}.",
            ),
            ("human", msg_template),
        ]
    )
    return prompt | llm | StrOutputParser()


def explain(inputs, ask_execute=True, ask_alternative=False):
    with Live(vertical_overflow="visible") as live:
        text = "🤖: "
        for chunk in explain_chain.stream(
            {
                **inputs,
                "lang_spec": (
                    f" in {settings.language} language" if settings.language else ""
                ),
            }
        ):
            text += chunk
            live.update(syntax(text))
    print()

    if not ask_execute:
        return
    print(f"🛠️ [bold green] {inputs['script']} [/bold green]")
    confirm = ask_confirm(explain=False, alternatives=ask_alternative)
    if confirm == "n":
        return "Command canceled by user", ToolMeta(
            stop_execution=True, skip_print=True
        )
    elif confirm == "c":
        pyperclip.copy(inputs["script"])
        return "Script copied to the clipboard!", ToolMeta(stop_execution=True)
    elif confirm == "a":
        return "alternative"



import typer
from typing import Optional, Annotated
from importlib.metadata import version
from shy_sh.agents.shy_agent.agent import ShyAgent
from shy_sh.settings import settings, configure_yaml
from shy_sh.agents.chains.explain import explain as do_explain
from shy_sh.utils import load_history
from rich import print
from time import strftime


def exec(
    prompt: Annotated[Optional[list[str]], typer.Argument(allow_dash=False)] = None,
    interactive: Annotated[
        Optional[bool],
        typer.Option(
            "-i",
            help="Interactive mode [default false if a prompt is passed]",
        ),
    ] = False,
    no_ask: Annotated[
        Optional[bool],
        typer.Option(
            "-x",
            help="Do not ask for confirmation before executing scripts",
        ),
    ] = False,
    explain: Annotated[
        Optional[bool],
        typer.Option(
            "-e",
            help="Explain the given shell command",
        ),
    ] = False,
    configure: Annotated[
        Optional[bool], typer.Option("--configure", help="Configure LLM")
    ] = False,
    display_version: Annotated[
        Optional[bool], typer.Option("--version", help="Show version")
    ] = False,
):
    if display_version:
        print(f"Version: {version(__package__ or 'shy-sh')}")
        return
    if configure:
        configure_yaml()
        return
    task = " ".join(prompt or [])
    print(f"[bold italic dark_orange]{settings.llm.provider} - {settings.llm.name}[/]")
    if explain:
        if not task:
            print("🚨 [bold red]No command provided[/]")
        do_explain(
            {
                "task": "explain this shell command",
                "script_type": "shell command",
                "script": task,
                "script_type": "shell command",
                "timestamp": strftime("%Y-%m-%d %H:%M:%S"),
            },
            ask_execute=False,
        )
        return

    if not task:
        interactive = True
    else:
        print(f"\n✨: {task}\n")
    try:
        ShyAgent(
            interactive=interactive,
            ask_before_execute=not no_ask,
        ).start(task)
    except Exception as e:
        print(f"🚨 [bold red]{e}[/bold red]")


def main():
    try:
        import readline

        readline.set_history_length(20)
    except Exception:
        pass
    load_history()
    typer.run(exec)


if __name__ == "__main__":
    main()



from rich import print
from shy_sh.agents.shy_agent.graph import shy_agent_graph
from shy_sh.agents.misc import get_graph_inputs, run_few_shot_examples
from shy_sh.utils import save_history
from langchain_core.messages import HumanMessage


class ShyAgent:
    def __init__(
        self,
        interactive=False,
        ask_before_execute=True,
    ):
        self.interactive = interactive
        self.ask_before_execute = ask_before_execute
        self.history = []
        self.examples = run_few_shot_examples()

    def _run(self, task: str):
        self.history.append(HumanMessage(content=task))
        inputs = get_graph_inputs(
            history=self.history,
            examples=self.examples,
            ask_before_execute=self.ask_before_execute,
        )

        res = shy_agent_graph.invoke(inputs)
        self.history += res["tool_history"]

    def start(self, task: str):
        if task:
            self._run(task)
        if self.interactive:
            new_task = input("\n✨: ")
            while new_task.endswith("\\"):
                new_task = new_task[:-1] + "\n" + input("    ")
            save_history()
            if new_task == "exit":
                print("\n🤖: 👋 Bye!\n")
                return

            self.start(new_task)



from langchain_core.messages import AIMessage
from shy_sh.settings import settings
from shy_sh.models import State
from shy_sh.agents.llms import get_llm_context
from shy_sh.utils import count_tokens, syntax
from shy_sh.agents.chains.shy_agent import shy_agent_chain
from shy_sh.agents.misc import has_tool_calls
from rich.live import Live

console_theme = {
    "lexer": "console",
    "theme": "one-dark",
    "background_color": "#181818",
}

loading_str = "⏱️ Loading..."


def chatbot(state: State):
    final_message = None
    history = _compress_history(state["history"], state["tool_history"])
    with Live(vertical_overflow="visible") as live:
        live.update(loading_str)
        for chunk in shy_agent_chain.stream({**state, "history": history}):
            final_message = chunk if final_message is None else final_message + chunk
            message = _parse_chunk_message(final_message)
            if _maybe_have_tool_calls(final_message, message):
                live.update(loading_str)
            else:
                live.update(
                    syntax(f"🤖: {message}"),
                    refresh=True,
                )
        message = _parse_chunk_message(final_message)
        ai_message = AIMessage(
            content=message, tool_calls=getattr(final_message, "tool_calls", [])
        )
        has_tools = has_tool_calls(ai_message)
        if not message or (settings.llm.agent_pattern == "react" and has_tools):
            live.update("")
        else:
            live.update(syntax(f"\n🤖: {message}"))
    return {"tool_history": [ai_message]}


def _maybe_have_tool_calls(message, parsed_message):
    return (
        not message.content
        or getattr(message, "tool_calls", None)
        or (parsed_message.startswith("{") and settings.llm.agent_pattern == "react")
    )


def _parse_chunk_message(chunk):
    if isinstance(chunk.content, list):
        return "".join(c.get("text") for c in chunk.content if c.get("type") == "text")
    else:
        return chunk.content


def _compress_history(history, tool_history):
    max_len = get_llm_context()
    tokens = count_tokens(history + tool_history)
    while tokens > max_len:
        history = history[2:]
        if not history:
            break
        tokens = count_tokens(history + tool_history)
    return history



import re
from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.messages import AIMessage
from shy_sh.agents.llms import get_llm
from shy_sh.settings import settings
from shy_sh.utils import detect_shell
from textwrap import dedent
from rich.live import Live


msg_template = dedent(
    """
    Find some alternative commands using different tools and variations of the original that can be used to accomplish the same task as the given command.

    Generate only one line of code for each alternative command.
    Do not suggest alternatives that are not compatible with the current shell or operating system.
    Sort the commands by relevance and usefulness, the most relevant and useful commands should be at the top.
    Output each command in a separate block of code respecting the following format:

    # Max 1 line description of the command before each command block{lang_spec}
    ```
    command
    ```

    Given command: 
    ```
    {cmd}
    ```
    """
)

example_sh = dedent(
    """
    # List all files in the current directory in long format
    ```
    ls -l
    ```

    # List all files in the current directory in long format including hidden files
    ```
    ls -la
    ```

    # Find all files in the current directory and its subdirectories including hidden files
    ```
    find . -type f
    ```

    # Find all files in the current directory and print the full path
    ```
    find . -type f -print
    ```
    """
)

example_cmd = dedent(
    """
    # List all files in the current directory in long format including hidden files
    ```
    dir /A /B
    ```

    # Find all files in the current directory and its subdirectories
    ```
    dir /S
    ```

    # Find all files in the current directory and its subdirectories using tree
    ```
    tree /F
    ```

    # Find all files in the current directory and its subdirectories in plan ASCII
    ```
    tree /F /A
    ```
    """
)

example_psh = dedent(
    """
    # List all files in the current directory including hidden files
    ```
    Get-ChildItem -Force
    ```

    # Find all files in the current directory and its subdirectories
    ```
    Get-ChildItem -Recurse
    ```

    # List all file names in the current directory and its subdirectories, including hidden files
    ```
    Get-ChildItem -Recurse -Force -Name
    ```

    # List all directories in the current directory
    ```
    Get-ChildItem -Directory
    ```
    """
)


def get_example():
    shell = detect_shell()
    if shell == "powershell":
        return example_psh
    if shell == "cmd":
        return example_cmd
    return example_sh


@chain
def alternative_commands_chain(inputs):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a helpfull shell assistant. The current date and time is {timestamp}.\nYou are running on {system} using {shell} as shell",
            ),
            (
                "human",
                msg_template.format(cmd="ls", lang_spec=""),
            ),
            AIMessage(content=get_example()),
            MessagesPlaceholder("history"),
            ("human", msg_template),
        ]
    )
    return prompt | llm | StrOutputParser()


def get_alternative_commands(inputs):
    with Live() as live:
        live.update("⏱️ Finding an alternative solutions...")
        response = alternative_commands_chain.invoke(
            {
                **inputs,
                "lang_spec": (
                    f" in {settings.language} language" if settings.language else ""
                ),
            }
        )
        live.update("")
    return re.findall(r"([^\n]+)\n```[^\n]*\n([^\n]+)\n```", response)



import re
import os
import platform
import subprocess
from typing import Literal
from tiktoken import get_encoding
from rich.prompt import Prompt
from rich.syntax import Syntax
from langchain_core.messages import HumanMessage, ToolMessage, AIMessage
from shy_sh.settings import settings

try:
    import readline
except Exception:
    readline = None


RL_HISTORY_FILE = os.path.expanduser("~/.config/shy/.history")


def load_history():
    if not readline:
        return
    try:
        readline.read_history_file(RL_HISTORY_FILE)
    except FileNotFoundError:
        pass


def save_history():
    if not readline:
        return
    readline.write_history_file(RL_HISTORY_FILE)


def clear_history():
    if not readline:
        return
    readline.clear_history()


def ask_confirm(explain=True, alternatives=False) -> Literal["y", "n", "c", "e", "a"]:
    clear_history()
    choices = ["n", "c", "no", "copy"]
    if explain:
        choices.extend(["e", "explain"])
    if alternatives:
        choices.extend(["a", "alternatives"])
    if not settings.safe_mode:
        choices.extend(["y", "yes"])

    ret = Prompt.ask(
        f"""\n [dark_orange]{ 
            'Do you need more details?' if settings.safe_mode else 'Do you want to execute this command?'
            }[/] [bold magenta][{
            '[underline]C[/]opy/[underline]n[/]o' if settings.safe_mode else '[underline]Y[/]es/[underline]n[/]o/[underline]c[/]opy'
            }{
            '/[underline]e[/]xplain' if explain else ''
            }{
            '/[underline]a[/]lternatives' if alternatives else ''
            }][/]""",
        choices=choices,
        default="c" if settings.safe_mode else "y",
        show_default=False,
        show_choices=False,
        case_sensitive=False,
    ).lower()[0]
    clear_history()
    load_history()
    return ret  # type: ignore


PRINT_THEMES = {
    "response": {
        "theme": "one-dark",
        "background_color": "#181818",
    },
    "command": {
        "background_color": "#212121",
    },
}


def syntax(text: str, lexer: str = "console", theme: str = "response"):
    return Syntax(
        text,
        lexer,
        word_wrap=True,
        **PRINT_THEMES.get(theme, {}),  # type: ignore
    )


def decode_output(process):
    try:
        response = process.stdout.decode() or process.stderr.decode()
    except UnicodeDecodeError:
        # windows
        import ctypes

        oemCP = ctypes.windll.kernel32.GetConsoleOutputCP()
        encoding = "cp" + str(oemCP)
        response = process.stdout.decode(encoding) or process.stderr.decode(encoding)
    return response


def decode_output2(text: bytes):
    try:
        response = text.decode()
    except UnicodeDecodeError:
        # windows
        import ctypes

        oemCP = ctypes.windll.kernel32.GetConsoleOutputCP()
        encoding = "cp" + str(oemCP)
        response = text.decode(encoding)
    return response


def run_shell(cmd: str):
    if cmd == "history" or cmd.startswith("history "):
        return get_shell_history()
    result = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    return decode_output(result)


def run_pty(cmd: str):
    import pty

    if cmd == "history" or cmd.startswith("history "):
        return get_shell_history()
    stdout = b""

    def read(fd):
        nonlocal stdout
        ret = os.read(fd, 1024)
        stdout += ret
        return ret

    ret_code = pty.spawn([detect_raw_shell(), "-c", cmd], read)
    return ret_code, decode_output2(stdout)


def stream_shell(cmd: str):
    if cmd == "history" or cmd.startswith("history "):
        return get_shell_history()
    result = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    while result.poll() is None:
        chunk = b""
        if result.stdout is not None and result.stdout.readable():
            chunk += result.stdout.read(1)
        yield decode_output2(chunk)
    remaining = b""
    if result.stdout is not None and result.stdout.readable():
        remaining += result.stdout.read()
    if result.stderr is not None and result.stderr.readable():
        remaining += result.stderr.read()

    if remaining:
        yield decode_output2(remaining)


def run_command(cmd: str):
    if detect_shell() in ["powershell", "cmd"]:
        result = ""
        for chunk in stream_shell(cmd):
            print(chunk, end="", flush=True)
            result += chunk
        result = result or "Exit code: 0"
    else:
        ret_code, result = run_pty(cmd)
        result = result or f"Exit code: {ret_code}"
    return result


def run_python(file: str):
    return run_command(f"python {file}")


def detect_raw_shell():
    return os.environ.get("SHELL") or os.environ.get("COMSPEC") or "sh"


def detect_shell():
    shell = detect_raw_shell()
    shell = shell.lower().split("/")[-1]
    if "powershell" in shell:
        return "powershell"
    elif "cmd" in shell:
        return "cmd"

    return shell


def detect_os():
    system = platform.system() or "linux"
    if system.lower() == "darwin":
        return "macos"
    return system


def count_tokens(
    messages: list, encoding_name: str = "o200k_base", offset: int = 2000
) -> int:
    text = "\n".join(msg.content for msg in messages)
    encoding = get_encoding(encoding_name)
    return len(encoding.encode(text)) + offset


def tools_to_human(messages):
    return [
        (
            HumanMessage(msg.content)
            if isinstance(msg, ToolMessage)
            else (
                AIMessage(msg.content or "tool_request")
                if isinstance(msg, AIMessage)
                else msg
            )
        )
        for msg in messages
    ]


HISTORY_FILES = {
    "bash": ".bash_history",
    "sh": ".bash_history",
    "zsh": ".zsh_history",
    "fish": ".local/share/fish/fish_history",
    "ksh": ".ksh_history",
    "tcsh": ".history",
}


def get_shell_history():
    try:
        shell = detect_shell()
        history_file = HISTORY_FILES[shell]
        with open(os.path.expanduser(f"~/{history_file}"), "r") as f:
            history = f.read()
        return "\n".join(
            [
                cmd
                for cmd in history.strip().split("\n")[:-1]
                if cmd != "shy"
                and not cmd.startswith("shy ")
                and ";shy " not in cmd
                and not cmd.endswith(";shy")
            ][-5:]
        )
    except Exception:
        return "I can't get the history for this shell"


def parse_code(code):
    code = re.sub(r"```[^\n]*\n", "", code)
    return code[: code.rfind("```")]
